#!/usr/bin/perl

# 'Know thyself' - locates itself and the name it was invoked as.

$CalledAs=$0;
$PWD=$ENV{"PWD"};

$MyDir='';
$MyDir=$1 if (${CalledAs}=~m#^(.+/)[^/]+#o);
if ($MyDir) {
	($MyName)=$CalledAs=~m#^${MyDir}(.+)$#o;
} else {
	$MyName=$CalledAs;
}
if (!($MyDir=~m#^/#o)) {
	$MyDir="${PWD}/${MyDir}";
}

chdir "$MyDir" || die ("Unable to locate own directory: $MyDir\n$!");

require('usenet-web-rc.pl');

foreach $newsgroup (sort keys(%newsgrouplist)) {
	&GetGroupInfo($newsgroup);
	if (!$ARCHIVEDIRECTORY || ($MessageCount eq '') || !$TRIPLEHEADERINDEX ||
		!$MESSAGEIDINDEX || !$IconPath || !$URLPath ||
		($UseArticleFile && ($HighMessage eq '')) ) {
		&InvalidResource;
		warn("Archiver failed - invalid resource: $newsgroup\n");
		next;
	}
	if ($UseActiveFile) {
	 	&ReadActiveFile;
	} else {
		&ReadNewsSpoolDirectory;
	}
	&ArchiveArticles;
	&PutGroupInfo($newsgroup);
	&UpdateResourceFile;
}

sub ArchiveArticles {
	local ($FILE,$Date,$Subject,$From,$MessageID);
	chdir "$NEWSSPOOL" || die ("Archiver was unable to open $NEWSSPOOL\n$!");
	foreach $FILE (@ArticleList) {
		chdir $NEWSSPOOL;
		if (-T $FILE) {
			if (!open(ARTICLEHANDLE,"$FILE")) {
				warn("Archiver unable to open $NEWSSPOOL/$FILE\n$1");
				next;
			}
			@ARTICLE=<ARTICLEHANDLE>;
			close(ARTICLEHANDLE);
			undef $Subject;
			undef $MessageID;
			undef $Date;
			undef $From;
			foreach $oline (@ARTICLE) {
				$line=$oline;
				chop $line;
				last if $line eq '';
				if ($line=~m/^(Subject: .+)$/o) {
					$Subject=$1;
				} elsif ($line=~m/^Message-I[Dd]: (<.+>)$/o) {
					$MessageID=$1;
				} elsif ($line=~m/^(Date: .+)$/o) {
					$Date=$1;
				} elsif ($line=~m/^(From: .+$)/o) {
					$From=$1;
				}
			}
			if (!$Subject || !$MessageID || !$Date || !$From) {
				warn("Archiver - $NEWSSPOOL/$FILE not a RFC 1036 article\n");
				next;
			}
			&MakeArchiveName($Date);
			$FullPath="${ARCHIVEDIRECTORY}/$ArchiveDatePath";
			$FullName="${FullPath}/${ARCHIVEFILENAME}";

			#Check if this article is already in archive
			next if (&IsDuplicate("$FullName"));
			$ARCHIVEFILENAME="${ARCHIVEFILENAME}.${uniqueid}";

			# Skip if unable to reach needed directory
			next if (!&MakePath("$FullPath"));
			
			# Skip if something went wrong saving the article
			next if (!&SaveArticle);

			# Update the index files
			$MessageCount++;
			&UpdateIndexes;
			if ($RemoveOriginal) {
				chdir $NEWSSPOOL;
				unlink $FILE;
			}
		}
	}
}

sub UpdateIndexes {

	chdir $ARCHIVEDIRECTORY;

	$IndexString=$MessageCount;
	$IndexString="0$MessageCount" if ($MessageCount < 1000000);
	$IndexString="00$MessageCount" if ($MessageCount < 100000);
	$IndexString="000$MessageCount" if ($MessageCount < 10000);
	$IndexString="0000$MessageCount" if ($MessageCount < 1000);
	$IndexString="00000$MessageCount" if ($MessageCount < 100);
	$IndexString="000000$MessageCount" if ($MessageCount < 10);

	if(!&DoIndexUpdate) {
		warn("Archiver Failed to update index file\n$!")
	}
	chdir "volume$year";
	if(!&DoIndexUpdate) {
		warn("Archiver Failed to update index file\n$!")
	}
	chdir "$month";
	if(!&DoIndexUpdate) {
		warn("Archiver Failed to update index file\n$!")
	}
}
sub DoIndexUpdate {
	local($RelName)="${ArchiveDatePath}/${ARCHIVEFILENAME}";
	return 0 if (!open(IDINDEX,">>$MESSAGEIDINDEX"));
	print IDINDEX "$IndexString $RelName $MessageID\n";
	close(IDINDEX);
	chmod 0644,"$MESSAGEIDINDEX";
	return 0 if (!open(ARCHIVEINDEX,">>$TRIPLEHEADERINDEX"));
	print ARCHIVEINDEX "${IndexString}:${RelName}:${Date}\n";
	print ARCHIVEINDEX "${IndexString}:${RelName}:${From}\n";
	print ARCHIVEINDEX "${IndexString}:${RelName}:${Subject}\n";
	close (ARCHIVEINDEX);
	chmod 0644,"$TRIPLEHEADERINDEX";
	1;
}

sub SaveArticle {
	chdir "$FullPath";

	if (-e "${ARCHIVEFILENAME}") {
		warn("Archiver Failed: File with same name already exists.\n$!");
		return 0;
	}
	if (!open(ARTICLEHANDLE,">${ARCHIVEFILENAME}")) {
		warn("Archiver Failed: Unable to write article to archive.\n$!");
		return 0;
	}
	print ARTICLEHANDLE @ARTICLE;
	close(ARTICLEHANDLE);
	chmod 0644,"$ARCHIVEFILENAME";
	system("$COMPRESS",$ARCHIVEFILENAME);
	1;
}


sub MakePath {
	local($PartialName)='';
	local($DirPath)="$_[0]";
	local(@PathComponents,$NewPartial);

	if (-e "${DirPath}" && -w _ && -d _) {
		return 1;
 	}

	chdir "/";
	@PathComponents=split(/\//,"${DirPath}");
	foreach $component (@PathComponents) {
		$NewPartial="${PartialName}${component}/";
		if (!(-e "$NewPartial")) {
			if (!(mkdir("$component",0755))) {
				warn("Archiver Failed: Unable to create directory. $!");
				return 0;
			}
			chmod 0755,"$component";
		}
		if (!(chdir ($NewPartial))) {
			warn("Archiver Failed: Unable to enter directory. $!");
			return 0;
		}
		$PartialName=$NewPartial;
	}
	1;
}

sub MakeArchiveName {
	local($EDate);
	
	$EDate=$_[0];
	$EDate=~s/^Date:\s+//o;
	$EDate=~s/^[A-Za-z]+, +//o;
	$day=''; $month=''; $year=''; $hour=''; $minute=''; $second='';
	($day,$month,$year,$hour,$minute)=$EDate=~m/^(\d\d?) +(\w\w\w)\w* +(\d\d+) +(\d\d):(\d\d)/o;
	($second)=$EDate=~m/^.*\d\d:\d\d:(\d\d)/o; # Bloody non RFC newsreaders...
	$second='00' if(!$second);
	$day=~s/^(\d)$/0$1/;
	$year+=1900 if (($year<1900) && ($year>60));
	$year+=2000 if ($year<100);
	$month=~tr/A-Z/a-z/;
	$ArchiveDatePath="volume${year}/${month}";
	$ARCHIVEFILENAME="${day}_${hour}${minute}${second}";	
}

sub IsDuplicate {
	local($BASEFILE)=$_[0];

	$uniqueid=0;
	$duplicate='';
	while(-e "${BASEFILE}.${uniqueid}.gz" && -r _) {
		if (!open (CHECKID,"gunzip -c ${BASEFILE}.${uniqueid}.gz |")) {
			warn("Unable to check ${BASEFILE}.${uniqueid}.gz for Message-ID - possible duplicate: $!");
			$uniqueid++;
			next;
		} else {
			@check=<CHECKID>;
			close(CHECKID);
			$match='';
			foreach $cline (@check) {
				chop $cline;
				last if (!$cline);
				($match)=$cline=~m/^Message-I[Dd]: +(<.+>)/o;
				$duplicate=($match eq $MessageID);
				last if	$duplicate;
				$uniqueid++;
			}
		last if $duplicate;
		}
	}
	return $duplicate;
}	
sub ReadNewsSpoolDirectory {
	opendir(SPOOLDIR,"$NEWSSPOOL") || die ("Archiver failed: unable to read newspool directory\n");
	@ArticleList=readdir(SPOOLDIR);
	closedir(SPOOLDIR);
}

sub ReadActiveFile {
	local ($name,$counter,$StartMessage)='';
	$StopMessage='';
	open(ACTIVE,"$ACTIVEFILE") || die ("Archiver failed: unable to open active file $ACTIVEFILE\n");
	while(<ACTIVE>) {
		($name,$StopMessage,$StartMessage)=m/^(.+)\s+([0-9]+)\s+([0-9]+)\s/o;
		last if ($name eq $newsgroup);
	}
	close(ACTIVE);
	$StartMessage=~s/^0+//;
	if ($name ne $newsgroup) {
		die ("Archiver failed: unable to find $newsgroup in active file $ACTIVEFILE\n");

	}
	if ($StartMessage eq '') {
		$StartMessage=0;
	}
	if ($StartMessage < $HighMessage) {
		$StartMessage=$HighMessage;
	}
	undef @ArticleList;

	if ($StopMessage > 0) {
		for ($counter=$StartMessage;$counter <= $StopMessage;$counter++) {
			push(@ArticleList,$counter);
		}
		$HighMessage=$StopMessage+1;
	}
}

sub InvalidResource {
        print 'The resource file entry for the newsgroup ',
		"'$newsgroup' was either did not exist or contained invalid ",
                " information.\n"
}

