#!/usr/bin/perl

# This is the fourth major rewrite of the Usenet-Web software
# Naturally, it is therefore version 1.0

# The Usenet-Web programs are copyrighted 1994 by Benjamin Franz
# (snowhare@netimages.com) and may be freely distributed and
# modified so long as no fees are charged.

# It would also be nice if you wrote me to tell me if you like it.

# This is the 'repair indexes' sub-program. It takes no
# parameters. It scans the entire archive to ensure no
# anomalies exist in the index files such as files present
# in the archive, but not the index or vice versa. It also
# rebuilds the index files sorted into strict date-time order.

# A secondary use is to convert Usenet 0.9 archives into
# the format required for the 1.0 program.

require('usenet-web-rc.pl');

&Initialize;
foreach $newsgroup (sort keys(%newsgrouplist)) {
	&GetGroupInfo($newsgroup);
	&ReadIndexFiles;
	&ReadArchiveDirectories;
	&ReconcileIndexes;
	&WriteIndexFiles;
	&PutGroupInfo($newsgroup);
	&UpdateResourceFile;
}

sub WriteIndexFiles {

	chdir ("${ARCHIVEDIRECTORY}");
	system("mv","${MESSAGEIDINDEX}","${MESSAGEIDINDEX}.old");
	system("mv","${TRIPLEHEADERINDEX}","${TRIPLEHEADERINDEX}.old");

	open (TOPLEVELID,">${ARCHIVEDIRECTORY}/${MESSAGEIDINDEX}");
	open (TOPLEVELTRIPLE,">${ARCHIVEDIRECTORY}/${TRIPLEHEADERINDEX}");

	$THISMONTH='';
	$THISYEAR='';
	$counter=1;

	foreach $key (sort keys(%DateOrderedIndex)) {
		$FilePath=$DateOrderedIndex{$key};
		$Subject=$IndexDataElements{"${FilePath}.Subject"};
		$Date=$IndexDataElements{"${FilePath}.Date"};
		$From=$IndexDataElements{"${FilePath}.From"};
		$MessageID=$IndexDataElements{"${FilePath}.MessageID"};
		$FilePath=~m#^(volume\d\d\d\d)/([a-z][a-z][a-z])/#o;
		$year=$1;
		$month=$2;

		if ($year ne $THISYEAR) {
			chdir ("${ARCHIVEDIRECTORY}/${year}");
			open (SECONDTIERID,">${MESSAGEIDINDEX}");
			open (SECONDTIERTRIPLE,">${TRIPLEHEADERINDEX}");
			$THISYEAR=$year;
		}
		if ($month ne $THISMONTH) {
			chdir ("${ARCHIVEDIRECTORY}/${year}/${month}");
			open (THIRDTIERID,">${MESSAGEIDINDEX}");
			open (THIRDTIERTRIPLE,">${TRIPLEHEADERINDEX}");
			$THISMONTH=$month;
		}

		$IndexNumber=&MakeIndexNumber($counter);
		print TOPLEVELID "${IndexNumber} ${FilePath} ${MessageID}\n";
		print TOPLEVELTRIPLE "${IndexNumber}:${FilePath}:Date: ${Date}\n",
			"${IndexNumber}:${FilePath}:From: ${From}\n",
			"${IndexNumber}:${FilePath}:Subject: ${Subject}\n";

		print SECONDTIERID "${IndexNumber} ${FilePath} ${MessageID}\n";
		print SECONDTIERTRIPLE "$IndexNumber:$FilePath:Date: ${Date}\n",
			"${IndexNumber}:${FilePath}:From: ${From}\n",
			"${IndexNumber}:${FilePath}:Subject: ${Subject}\n";

		print THIRDTIERID "${IndexNumber} ${FilePath} ${MessageID}\n";
		print THIRDTIERTRIPLE "${IndexNumber}:${FilePath}:Date: ${Date}\n",
			"${IndexNumber}:${FilePath}:From: ${From}\n",
			"${IndexNumber}:${FilePath}:Subject: ${Subject}\n";
		$counter++;
	}
	$counter--;
	$MessageCount=$counter;
	close(TOPLEVELID);
	close(TOPLEVELTRIPLE);
	close(SECONDTIERID);
	close(SECONDTIERTRIPLE);
	close(THIRDTIERID);
	close(THIRDTIERTRIPLE);
	chmod 0644,"${ARCHIVEDIRECTORY}/${MESSAGEIDINDEX}";
	chmod 0644,"${ARCHIVEDIRECTORY}/${TRIPLEHEADERINDEX}";
}

sub ReconcileIndexes {
	foreach $key (keys(%DateOrderedIndex)) {
		$IndexNumber=$DateOrderedIndex{$key};

		if (!$IndexDataElements{"${IndexNumber}.Date"} ||
			!$IndexDataElements{"${IndexNumber}.From"} ||
			!$IndexDataElements{"${IndexNumber}.Subject"} ||
			!$IndexDataElements{"${IndexNumber}.MessageID"} ||
			!$IndexDataElements{"${IndexNumber}.directory"}) {

			if (!$IndexDataElements{"${IndexNumber}.directory"}) {
				undef($DateOrderedIndex{$key});
				print "No file found in archive matching index entry.\n",
					"Unmatched Index Entry:",
					"\n    Date: ",
					$IndexDataElements{"${IndexNumber}.Date"},
					"\n    From: ",
					$IndexDataElements{"${IndexNumber}.From"},
					"\n    Subject: ",
					$IndexDataElements{"${IndexNumber}.Subject"},
					"\n    Message-ID: ",
					$IndexDataElements{"${IndexNumber}.MessageID"},
					"\nEntry was removed from index.\n\n";
				next;
			}
			if(!&ReadFileHeaders("$DateOrderedIndex{$key}")) {
				undef($DateOrderedIndex{$key});
				print "${ARCHIVEDIRECTORY}/",
					$DateOrderedIndex{$key},
					" was not a valid RFC-1036 compliant ",
					"article.\nEntry was NOT included in index.\n",
					"You should check this file for damage.\n\n";
				next;
			}
		}
	}
}

sub ReadFileHeaders {
	local($thisarticle)=$_[0];
	local($from,$subject,$date,$messageid)='';

	chdir "${ARCHIVEDIRECTORY}";

        if (-e "$thisarticle.gz" && -r _) {
              $thisarticle = "$UNCOMPRESS ${thisarticle}.gz |";
        } elsif (-e "$thisarticle.Z" && -r _)  {
              $thisarticle = "$UNCOMPRESS ${thisarticle}.Z |";
        } elsif (-e "$thisarticle.z" && -r _)  {
              $thisarticle = "$UNCOMPRESS ${thisarticle}.z |";
        }

        if (!open (ARTICLEHANDLE,"$thisarticle")) {
                return 0;
        }

        while ($line=<ARTICLEHANDLE>) {
                chop $line;
		last if (!$line);

		if ($line=~m#^Subject: (.+)#o) {
			$subject=$1;
			next;
		} elsif ($line=~m#^Date: (.+)#o) {
			$date=$1;
			next;
		} elsif ($line=~m#^From: (.+)#o) {
			$from=$1;
			next;
		} elsif ($line=~m#^Message-[Ii][Dd]: (<.+>)#o) {
			$messageid=$1;
			next;
		}
	}
	close (ARTICLEHANDLE);

	$success=($messageid && $from && $date && $subject);
	return 0 if (!$success);	
	$IndexDataElements{"${IndexNumber}.Date"}=$date;
	$IndexDataElements{"${IndexNumber}.From"}=$from;
	$IndexDataElements{"${IndexNumber}.Subject"}=$subject;
	$IndexDataElements{"${IndexNumber}.MessageID"}=$messageid;
	1;
}

sub MakeIndexNumber {
	local($IndexNumber)=$_[0];

	$IndexString=$IndexNumber;
	$IndexString="0$IndexNumber" if ($IndexNumber < 1000000);
	$IndexString="00$IndexNumber" if ($IndexNumber < 100000);
	$IndexString="000$IndexNumber" if ($IndexNumber < 10000);
	$IndexString="0000$IndexNumber" if ($IndexNumber < 1000);
	$IndexString="00000$IndexNumber" if ($IndexNumber < 100);
	$IndexString="000000$IndexNumber" if ($IndexNumber < 10);
	return $IndexString;
}
sub ReadArchiveDirectories {
	local($message,$MONTH,$VOLUME,$Year,@RawMonthList,$FilePath);

	undef @RawYearList;
	undef %MonthDirs;
	undef %YearsDirs;

	opendir(TOPLEVEL,"${ARCHIVEDIRECTORY}") ||
		die ("Unable to open ${ARCHIVEDIRECTORY}\n$!");
	@RawYearList=grep(/^volume\d\d\d\d$/,readdir(TOPLEVEL));
	closedir(TOPLEVEL);
	foreach $VOLUME (@RawYearList) {
		if (!opendir(SECONDTIER,"${ARCHIVEDIRECTORY}/${VOLUME}")) {
			die ("Unable to open ${ARCHIVEDIRECTORY}/${VOLUME}\n$!");
		}
		($Year)=$VOLUME=~m/volume(\d\d\d\d)/;
		$YearDirs{$year}=$VOLUME;
		@RawMonthList=grep(/^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)$/,readdir(SECONDTIER));
		closedir(SECONDTIER);
		foreach $MONTH (@RawMonthList) {
			$MonthDirs{$MONTH}="${VOLUME}/${MONTH}";
			if (!opendir(THIRDTIER,"${ARCHIVEDIRECTORY}/${VOLUME}/${MONTH}")) {
				die ("Unable to open ${ARCHIVEDIRECTORY}/${VOLUME}/${MONTH}\n$!");
			}
			@RawDayList=grep(/\d+_\d+\.\d+/,readdir(THIRDTIER));
			closedir(THIRDTIER);
			$DateOrder="${Year}.$MonthNumber{$MONTH}.";
			foreach $message (@RawDayList) {
				$message=~s/\.gz$//oi;
				$FilePath="${VOLUME}/${MONTH}/${message}";
				$DateOrderedIndex{"${DateOrder}${message}"}="$FilePath";
				$IndexDataElements{"${FilePath}.directory"}='y';
			}
		}
	}
}

sub ReadIdIndexFiles {
	local(@IdIndexes);
	local($year,$month,$day,$time,$FilePath,$MessageId,$DateOrder)='';
	if (!open(IDINDEX,"$ARCHIVEDIRECTORY/$MESSAGEIDINDEX")) {
		warn("Unable to read master $MESSAGEIDINDEX file\n$!");
		return 0;
	}
	@IdIndexes=<IDINDEX>;
	close(IDINDEX);
	$MessageId='';
	$FilePath='';
	chop @IdIndexes;
	while(@IdIndexes) {
		$line=pop(@IdIndexes);
		$line=~s#^\s+$##o;
		next if (!$line);
		($FilePath,$MessageId)=$line=~m#^\d+ (volume\d+/\w+/\d+_\d+\.\d+) (<.+>)$#o;
		if (!$FilePath || !$MessageId) {
			print " Invalid $MESSAGEIDINDEX Entry: $line\n";
			next;
		}
		$IndexDataElements{"${FilePath}.MessageID"}=$MessageId;
		($year,$month,$day,$timestring)=$FilePath=~m#volume(\d+)/(\w+)/(\d+)_(\d+\.\d+)#o;
		$DateOrder="${year}.$MonthNumber{$month}.${day}_${timestring}";
		$DateOrderedIndex{"$DateOrder"}=$FilePath;
	}
}

sub ReadTripleIndexFiles {
	local(@TripleIndexes);
	local($year,$month,$day,$time,$FilePath,$Type,$Data,$DateOrder)='';
	if (!open(TRIPLEINDEX,"$ARCHIVEDIRECTORY/$TRIPLEHEADERINDEX")) {
		warn("Unable to read master $TRIPLEHEADERINDEX file\n$!");
		return 0;
	}
	@TripleIndexes=<TRIPLEINDEX>;
	close(TRIPLEINDEX);
	chop @TripleIndexes;
	while(@TripleIndexes) {
		$line=pop(@TripleIndexes);
		$line=~s#^\s+$##o;
		next if (!$line);
		($FilePath,$Type,$Data)=$line=~m#^\d+:(volume\d+/\w+/\d\d_\d+\.\d+):(From|Date|Subject): (.*)$#o;
		if (!$FilePath || !$Type || !$Data) {
			print " Invalid $TRIPLEHEADERINDEX Entry: $line\n";
			next;
		}
		$IndexDataElements{"${FilePath}.${Type}"}=$Data;
		($year,$month,$day,$timestring)=$FilePath=~m#volume(\d+)/(\w+)/(\d+)_(\d+\.\d+)#o;
		$DateOrder="${year}.$MonthNumber{$month}.${day}_${timestring}";
		$DateOrderedIndex{"$DateOrder"}=$FilePath;
	}
}

sub ReadIndexFiles {
	undef %DateOrderedIndex;
	undef %IndexDataElements;
	&ReadIdIndexFiles;
	&ReadTripleIndexFiles;
}

sub Initialize {
	%MonthNumber=('jan','01',
			'feb','02',
			'mar','03',
			'apr','04',
			'may','05',
			'jun','06',
			'jul','07',
			'aug','08',
			'sep','09',
			'oct','10',
			'nov','11',
			'dec','12',
	);
}
